/*  ----------------------------------------------------------------<Prolog>-
    Name:       com.imatix.ServletRunner.HttpServletRequestImpl
    Title:      Implementation of javax.servlet.http.HttpServletRequest
    Package:    iMatix Servlet Runner

    Written:    99/01/11  Ewen McNeill <ewen@imatix.com>
    Revised:    99/03/19

    Copyright:  Copyright (c) 1991-99 iMatix Corporation
    License:    This is free software; you can redistribute it and/or modify
                it under the terms of the SMT License Agreement as provided
                in the file LICENSE.TXT.  This software is distributed in
                the hope that it will be useful, but without any warranty.
 ------------------------------------------------------------------</Prolog>-*/

package com.imatix.ServletRunner;

import java.io.ByteArrayInputStream;
import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.io.Reader;
import java.io.IOException;
import java.text.DateFormat;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;
import javax.servlet.ServletInputStream;
import javax.servlet.http.*;

/**
    iMatix Servlet Runner implementation of 
    javax.servlet.http.HttpServletRequest.
 */
// XXX: Sessions are only marked as accessed by client if the servlet
//      attempts to retrieve the session.
public class HttpServletRequestImpl implements HttpServletRequest
{
    private static final String SESSION_COOKIE_NAME = "SessionID";

    private Hashtable CGI_Environment;
    private byte[]    postedData;

    private Hashtable servletParams;

    private ByteArrayInputStream is;
    private ServletInputStream   sis;
    private BufferedReader       brd;

    private HttpSessionContextImpl  sessionContext;
    private HttpServletResponseImpl response;

    HttpServletRequestImpl (Hashtable requestContext, byte[] requestInput,
                            HttpSessionContextImpl servletSessionContext,
			    HttpServletResponseImpl servletResponse)
    {
        CGI_Environment = requestContext;
	postedData      = requestInput;
	sessionContext  = servletSessionContext;
	response        = servletResponse;
    }

    // The following functions are in the HttpServletRequest interface

    /**
        Gets the cookies found in this request.

	Cookies are passed in the Cookie: HTTP header, which turns up
	in our environment as HTTP_COOKIE.  This header main contain
	multiple cookies, and may be in the original Netscape format,
	or in RFC2109 format.  If RFC2109 format cookies are found
	then the additional attributes (where present) will be set on
	the cookies; otherwise this attributes will be left as their
	default.

	XXX: Will not handle parsing cookies with ";" or "," in them
	correctly at present.
     */
    public Cookie[] getCookies()
    {
        Vector cookies = new Vector();
        String cookieList = (String)CGI_Environment.get("HTTP_COOKIE");
	int cookieVersion = 0;

	if (cookieList != null) 
	{
	    // Parse cookies string, looking for name=value pairs.  If it is
	    // RFC2109 then the first will be $Version, and any other
	    // $... names will be special.  We save the version for
	    // later, and set on each cookie (if non-zero).
	    // Cookies are separated by either ";" or "," (RFC2109 says
	    // ";" separates attributes of cookie, "," separates cookie).
	    //
	    // XXX: We need to handle ";" and "," inside the quotes as
	    // being part of the attribute/value; currently parsing will
	    // go wrong.
	    //
	    // XXX: Should parse cookies once and then keep array around
	    // to pass out/clone.
	    //
	    // XXX: HTTP_COOKIE2 appears to contain version string, if
	    // any.
	    //
	    final int attributeSep = (int)';';
	    final int cookieSep    = (int)',';
	    final int equals       = (int)'=';
	    final int length = cookieList.length();
	    int start = 0;

	    while (start < length)
	    {
	        // Get Name of pair
                int end = cookieList.indexOf(equals, start);

		String name = cookieList.substring(start, end);
		start = end + 1;

                // Find value of pair
                int nextAttSep = cookieList.indexOf(attributeSep, start);
		int nextCookieSep = cookieList.indexOf(cookieSep, start);

		if (nextCookieSep > start && nextCookieSep < nextAttSep)
		    end = nextCookieSep;
		else
		    end = nextAttSep;

                if (end < 0)
		    end = length;

                String value = cookieList.substring(start, end);
		start = end + 1;

		// Tidy up the name that we retrieved
		name = name.trim();

		// Tidy up the value that we retrieved (trim off
		// separator, strip double quotes)
		//
		value = value.trim();
		if (value.endsWith(";") || value.endsWith(","))
		    value = value.substring(0, value.length() - 1);

                if (value.startsWith("\"") && value.endsWith("\""))
		    value = value.substring(1, value.length() - 1);

		// Figure out if this is RFC2109 attribute ($....)
		if (name.startsWith("$"))
		{
		    if (name.equalsIgnoreCase("$Version"))
		    {
		        try {
		            cookieVersion = Integer.parseInt(value);
			}
			catch (NumberFormatException e) {}
		    }
		    else if (name.equalsIgnoreCase("$Path"))
		    {
		        Cookie last = (Cookie)cookies.lastElement();
			if (last != null)
			    last.setPath(value);
		    }
		    else if (name.equalsIgnoreCase("$Domain"))
		    {
		        Cookie last = (Cookie)cookies.lastElement();
			if (last != null)
			    last.setDomain(value);
		    }
		}
		else
		{
		    // Nope, it is a new cookie.  Let's create one.

                    Cookie cookie = new Cookie(name, value);
		    if (cookieVersion > 0)
		        cookie.setVersion(cookieVersion);
		    cookies.addElement(cookie);
		}
	    }
	}

        // Pull cookies out into an array
        int noCookies = cookies.size();
	Cookie cookieArray [] = new Cookie[noCookies];
	cookies.copyInto(cookieArray);

	return cookieArray;
    }

    /**
        Gets the HTTP method (eg, GET, POST, PUT, etc) that was used
	to make this request.  The value of the CGI variable
	REQUEST_METHOD is returned.
     */
    public String getMethod()
    {
        return (String)CGI_Environment.get("REQUEST_METHOD");
    }

    /**
       Gets the part of the request URI that is to the left of the query
       string. 

       This is a concatenation of the CGI variable SCRIPT_NAME and the 
       CGI variable PATH_INFO (untranslated).
     */
    public String getRequestURI()
    {
        String servletPath = getServletPath();
	String pathInfo    = getPathInfo();

	if (pathInfo != null)
	    return servletPath + "/" + pathInfo;
        else
	    return servletPath;
    }

    /**
        Gets the part of the request URI that refers to the servlet
	being invoked.  The value of the CGI variable SCRIPT_NAME
	is returned.
     */
    public String getServletPath()
    {
        return (String)CGI_Environment.get("SCRIPT_NAME");
    }

    /**
        Returns any optional extra path information following the
	servlet path of the request's URI, which precedes the query
	string.  The value of the CGI variable PATH_INFO is returned.
     */
    public String getPathInfo()
    {
        return (String)CGI_Environment.get("PATH_INFO");
    }

    /**
        Returns any optional extra path information following the
	servlet path of the requests URI, which precedes the query
	string, after translating it to a real path.  For now, returns
	CGI PATH_TRANSLATED.
     */
    public String getPathTranslated()
    {
        return (String)CGI_Environment.get("PATH_TRANSLATED");
    }

    /**
        Returns the query string portion of the requested URI.  The
	value of the CGI variable QUERY_STRING is returned.
     */
    public String getQueryString()
    {
        return (String)CGI_Environment.get("QUERY_STRING");
    }

    /**
        Gets the name of the user making this request, as set with the
	HTTP authentication.  The value of the CGI variable REMOTE_USER
	is returned.
     */
    public String getRemoteUser()
    {
        return (String)CGI_Environment.get("REMOTE_USER");
    }

    /**
        Gets the authentication scheme used for this request.  The
	value of the CGI variable AUTH_TYPE is returned.
     */
    public String getAuthType()
    {
        return (String)CGI_Environment.get("AUTH_TYPE");
    }

    /**
        Retrieves the named header as a String.

	The header strings are copied into the environment as
	HTTP_<headername>, with the headername in upper case.
	So we force the name into upper case, and then prefix
	the HTTP_, then try to retrieve it from the environment.
     */
    public String getHeader(String name)
    {
    	String envVarName = "HTTP_" + name.toUpperCase();
        return (String)CGI_Environment.get(envVarName);
    }

    /**
        Retrieves the value from a named integer header, as an int.
	If the field cannot be found, then -1 is returned.

        @see getHeader
	@exception NumberFormatException If the field does not contain
	                                 a number.
     */
    public int getIntHeader(String name) throws NumberFormatException
    {
        String headerField = getHeader(name);

	if (headerField == null)
	    return -1;

	return Integer.parseInt(headerField);
    }

    /**
        Retrieves the value from a named date header, as a long.
	The date is parsed using the DateFormat.parse() method, 
	for the default DateFormat.
 
        @return The value of the date field, or -1 if the field
	        is not found.

        @see getHeader
	@exception IllegalArgumentException  Thrown if the field cannot
	           be parsed as a Date.

     */
    public long getDateHeader(String name) throws IllegalArgumentException
    {
        String headerField = getHeader(name);

	if (headerField == null)
	    return (long)-1;

        try {
  	    return DateFormat.getDateInstance().parse(headerField).getTime();
	}
	catch (java.text.ParseException e)
	{
	    throw new IllegalArgumentException("Field does not contain date");
	}
    }

    /**
        Retrieves a list of header names accompanying this request.

	The header names are found in the environment as HTTP_<header>.
	We parse the environment to find those header names, put them
	into a vector, and return an enumeration of that.  Those
	names can be used with getHeader, etc.

	@see getHeader
     */
    public Enumeration getHeaderNames()
    {
        Enumeration envEnum     = CGI_Environment.keys();
	Vector      headerNames = new Vector();

	while (envEnum.hasMoreElements())
	{
	    String envName = (String)envEnum.nextElement();

	    if (envName.startsWith("HTTP_"))
	        headerNames.addElement(envName.substring(5));
	}

        return headerNames.elements();
    }

    /**
        Either create a new session (if create is true and it is
        necessary), or retrieve an existing session for this request.

        XXX: Currently only handles sessions from cookies.
     */
    public HttpSession getSession(boolean create)
    {
	String sessionID = getRequestedSessionId();

	if (sessionID != null)
	{
	    HttpSessionImpl session = 
	        (HttpSessionImpl)sessionContext.getSession(sessionID);
	    if (session != null)
	    {
	        session.recordSessionAccess();       // Mark session access
	        return session;
            }
	}

        // If we get through to here, either we have no session ID, or 
	// there is no session for that ID.  If we can, create a new
	// session ID.  (The old sessionID, if any, is invalid, and will
	// be overwritten by the new cookie.)
	//
        if (create)
        {
	    HttpSessionImpl session = new HttpSessionImpl(sessionContext);

	    // Add to context, and outgoing response
	    sessionContext.addSession(session);
            response.addCookie(new Cookie(SESSION_COOKIE_NAME, 
	                                  session.getId()));

	    return session;
	}

	// If we fall through to here, no existing session, and we are
	// not to create a session.
        return null;
    }

    /**
        Retrieves the requested session Id for this session.  This
        may differ from the actual session Id.  Returns null if no
	session was requested.

	XXX: Currently only handles sessions from cookies.
     */
    public String getRequestedSessionId()
    {
        Cookie cookies[] = getCookies(); 

        if (cookies != null)
	{
           for (int i = 0; i < cookies.length; i++)
   	       if (cookies[i].getName().equals(SESSION_COOKIE_NAME))
	           return cookies[i].getValue();
        }

	// If we fall through to here no cookie requesting session
        return null;
    }

    /**
        Returns true if the session ID is valid.

	Requested session ID is considered valid if we can retrieve
	a session based on it.
     */
    public boolean isRequestedSessionIdValid()
    {
        return (getSession(false) != null);
    }

    /**
        Returns true if the session ID is derived from a Cookie.

	XXX: Only cookie-based sessions implemented at present
     */
    public boolean isRequestedSessionIdFromCookie()
    {
        return true;
    }

    /**
        Returns true if the session ID is derived from a URL.

	XXX: Only cookie-based sessions implemented at present
     */
    public boolean isRequestedSessionIdFromUrl()
    {
        return false;
    }

    // These functions are in the ServletRequest interface, which
    // is extended by the HttpServletRequest interface

    /**
        Returns the size of the request entity data, or -1 if not
	known.  The value of the CGI variable CONTENT_LENGTH is
	returned, if present and convertable, else -1.
     */
    public int getContentLength ()
    {
       try {
           return Integer.parseInt (
	                     ((String)CGI_Environment.get("CONTENT_LENGTH")));
       }
       catch (Exception e) {}

       return -1;
    }

    /**
        Returns the MIME media type of the request entity data, or
	null if not known.  The value of the CGI variable
	CONTENT_TYPE is returned.
     */
    public String getContentType()
    {
        return (String)CGI_Environment.get("CONTENT_TYPE"); 
    }

    /**
        Returns the protocol and version of the request, as a string
	in the form <protocol>/<major version>.<minor version>.  The
	value of the CGI variable SERVER_PROTOCOL is returned.
     */
    public String getProtocol()
    {
        return (String)CGI_Environment.get("SERVER_PROTOCOL");
    }

    /**
        Returns the scheme of the URL used in the request (http, https,
	etc).  This can be used in reconstructing the URL. 

        This is extracted from the SERVER_URL (up to the ":") if that 
	CGI environment variable is present.
     */
    public String getScheme()
    {
        String serverURL = (String)CGI_Environment.get("SERVER_URI");

	if (serverURL != null)
	{
	    int firstColon = serverURL.indexOf((int)':');

	    if (firstColon > 0)
	        return serverURL.substring(0,firstColon);
	}

        // Default scheme, if we cannot determine it
	return "http";
    }

    /**
        Returns the host name of the server that received the request.
	The value of the CGI variable SERVER_NAME is returned.
     */
    public String getServerName()
    {
        return (String)CGI_Environment.get("SERVER_NAME");
    }

    /**
        Returns the port number of the server that received the request.
	The value of the CGI variable SERVER_PORT is returned.
     */
    public int getServerPort()
    {
        return Integer.parseInt ((String)CGI_Environment.get("SERVER_PORT"));
    }

    /**
        Returns the IP address of the agent that sent the request.
	The value of the CGI variable REMOTE_ADDR is returned.
     */
    public String getRemoteAddr()
    {
        return (String)CGI_Environment.get("REMOTE_ADDR");
    }

    /**
        Returns the fully qualified host name of the agent that
	sent the request.  The value of the CGI variable REMOTE_HOST
	is returned.
     */
    public String getRemoteHost()
    {
        return (String)CGI_Environment.get("REMTOE_HOST");
    }

    /**
        Returns the real path after applying alias expansions to
	the specified virtual path.  If "/" is given as an argument, the
	document root is returned.

	XXX: Note: Not implemented yet.  Not sure how we will do alias
	expansions (if at all).
     */
    public String getRealPath(String virtualPath)
    {
        return null;
    }

    /**
        Returns an input stream for reading binary data in the request
	body.  Since our input data is in a byte array, we use a
	ByteArrayInputStream to handle reading it.  An IllegalStateException 
	will be thrown if getReader() has already been called, but we
	try to handle multiple calls to getInputStream.

	@exception IllegalStateException If getReader() has already been
	                                 called on this object.
     */
    public ServletInputStream getInputStream() throws IOException
    {
        if (sis != null)
	    return sis;          // Return existing stream
	
	if (is != null)
	    throw new IllegalStateException(
 	                     "Call to getInputStream() after getReader()");

        // A New request -- open the appropriate things and return them.
	is = new ByteArrayInputStream(postedData);
	sis = new ServletInputStreamImpl(is);
	return sis;
    }

    /**
        Returns a string containing the (single) value for the specified
	parameter, or null if the parameter does not exist.  Since this
	is an HTTP Servlet, the parameters are taken out of the
	QUERY_STRING or posted data, as suggested by the ServletRequest 
	interface.

	Implementation defined behaviour; if there is more than one
	value that matches a given parameter, then the first value 
	found in the hash will be returned.
     */
    public String getParameter(String name)
    {
        parseServletParameters();

        try {
	    return ((String[])servletParams.get(name))[0];
	}
	catch (NullPointerException e)
	{
	    return null;
	}
    }

    /**
        Returns an array of strings containing the values for the 
	specified parameter, or null if the parameter does not exist.
	Since this is an HTTP Servlet, the parameters are taken out of
	the QUERY_STRING or posted data, as suggested by the
	ServletRequest interface
     */
    public String[] getParameterValues(String name)
    {
        parseServletParameters();

	if (servletParams != null)
            return (String[])servletParams.get(name);
	else
	    return null;
    }

    /**
        Returns an enumeration of the parameter names.  The parameters
	are taken from the QUERY_STRING or posted data.
     */
    public Enumeration getParameterNames()
    {
        parseServletParameters();

	if (servletParams != null)
	    return servletParams.keys();
	else 
            return new Hashtable().keys();     // Avoid returning null
    }

    /**
        Returns the value of the named attribute of the request, or
	null if the attribute does not exist. 

	XXX: Note: Not implemented yet.  (This one could be tricky)
     */
    public Object getAttribute(String name)
    {
        return null;
    }

    /**
        Returns a buffered reader for reading text in the request
	body.  This translates character set encodings as appropriate.
	Since the request body is in a byte array we use a
	ByteArrayInputStream to handle reading the array.

	@exception IllegalStateException Thrown if getInputStream() has
	                                 been called on this object.
     */
    public BufferedReader getReader() throws IOException
    {
        if (brd != null)
	    return brd;               // Return existing reader

        if (is != null)
	    throw new IllegalStateException(
	              "Call to getReader after getInputStream");

        // New request -- open the appropriate things and return them
	is  = new ByteArrayInputStream(postedData);
	brd = new BufferedReader(new InputStreamReader(is));
	return brd;
    }

    /**
        Returns the character set encoding used for the input of this
	request.  Since CGI doesn't support this concept, the value has
	to be hard coded for now.

	XXX: Note: Not implemented yet.
     */
    public String getCharacterEncoding()
    {
        return null;
    }

    // Private implementation functions

    /** 
        Parse the parameters to the servlet, out of both the query
	string and also the posted data (if any).  The input stream
	is read from the point where the servlet is up to, if it has
	started reading (this is consistent with the API documentation),
	and will be ignored if an IllegalStateException is thrown.
        Generally the input stream will be valid only for POST/PUT
	requests at present, but for consistency we parse it now
	always.

	The returned data structure is a Hash of parameter names, 
	which map to arrays of Strings which hold the values.  We
	join the values returned from the query string and from the
	posted data together into a suitable format.
      */
    private synchronized void parseServletParameters()
    {
        if (servletParams != null)
	    return;                       // Been there, done that

	Hashtable queryStringParams = null;
	Hashtable inputStreamParams = null;

        String queryString = getQueryString();
	if (queryString != null)
	    queryStringParams = HttpUtils.parseQueryString(queryString);

        try {
	    // XXX: Can we get a better estimate of the length than this?
	    // Eg, if the client has actually read some data already?
	    int inputStreamLength = getContentLength();
	    if (inputStreamLength > 0)
	    {
	        ServletInputStream inputStream = getInputStream();

	        inputStreamParams = 
		     HttpUtils.parsePostData(inputStreamLength, inputStream);
            }
	}
	catch (IllegalStateException e) {}
	catch (IOException e)           {}

	if (queryStringParams == null && inputStreamParams == null)
	    return;                        // No parameters available

        if (queryStringParams == null) //inputStreamParams != null
	{
	    // It appears we have to convert the Hashtable from one that 
	    // contains both Strings and arrays of Strings to one that 
	    // contains only arrays of Strings (in which case the interface 
	    // is bogus).  If not, this should still be safe.

	    servletParams = new Hashtable();  // safe: synchronized; gatekeeper
	    Enumeration isParamsEnum = inputStreamParams.keys();
	    while (isParamsEnum.hasMoreElements())
	    {
	        String entryName = (String)isParamsEnum.nextElement();
		Object entryValue = inputStreamParams.get(entryName);
		if (entryValue instanceof String)
		{
		    String[] arrayOfOne = new String[1];
		    arrayOfOne[0] = (String)entryValue;
		    servletParams.put(entryName, arrayOfOne);
		}
		else  // if not a String should be an array....
		{
		    servletParams.put(entryName, entryValue);
		}
	    }

	    return;
	}

	if (inputStreamParams == null) //queryStringParams != null
	{
            servletParams = queryStringParams;
	    return;
	}

	// Otherwise we have two hashtables and we have to combine their
	// values.  Since the queryStringParams one is in the right
	// format, we start with that one first.
	servletParams = queryStringParams;    // safe: synchronized; gatekeeper

        Enumeration isParamsEnum = inputStreamParams.keys();
        while (isParamsEnum.hasMoreElements())
        {
            String entryName = (String)isParamsEnum.nextElement();
            Object entryValue = inputStreamParams.get(entryName);

	    // There are four possibilities:
	    // 1. The hashtable already has an entry for this name
	    //    in it, and what we have is a String; in which
	    //    case we need to make a new array with one more string
	    //    in it.
	    // 2. The hashtable already has an entry for this name,
	    //    and the entry is an arary, in which case we need
	    //    to join the two arrays together.
	    // 3. There is no entry for this name, and we have a string,
	    //    in which case we need to build an array of one.
	    // 4. There is no entry for this name, and we have an array
	    //    in which case we can use it as is.

            if (servletParams.containsKey(entryName))// Possibility 1 or 2
	    {  
	       if (entryValue instanceof String)     // Possibility 1
	       {  
	           String[] existing = 
		       ((String[])servletParams.get(entryName));
	           String[] temp = new String[existing.length + 1];
                   System.arraycopy(existing, 0, temp, 0, existing.length);
                   temp[temp.length-1] = (String)entryValue;
		   servletParams.put(entryName, temp);
	       }
	       else                                  // Possibility 2
	       {
	           String[] existing = 
		       ((String[])servletParams.get(entryName));
	           String[] toAdd = (String[])entryValue;
	           String[] temp = new String[existing.length + toAdd.length];
                   System.arraycopy(existing, 0, temp, 0, existing.length);
		   System.arraycopy(toAdd, 0, 
		                    temp, existing.length, toAdd.length);
		   servletParams.put(entryName, temp);
	       }
	    }
	    else                                     // Possiblity 3 or 4
	    {
		if (entryValue instanceof String)    // Possiblity 3
		{
		    String[] arrayOfOne = new String[1];
		    arrayOfOne[0] = (String)entryValue;
		    servletParams.put(entryName, arrayOfOne);
		}   
		else                                 // Possiblity 4
		{
		    servletParams.put(entryName, entryValue);
		}   
	    }
        } // while entries in inputStreamParameters
    } // parseServletParameters();
}
