/*  ----------------------------------------------------------------<Prolog>-
    Name:       com.imatix.ServletRunner.HttpServletResponseImpl
    Title:      Implementation of javax.servlet.http.HttpServletResponse
    Package:    iMatix Servlet Runner

    Written:    99/01/11  Ewen McNeill <ewen@imatix.com>
    Revised:    99/01/17

    Copyright:  Copyright (c) 1991-99 iMatix Corporation
    License:    This is free software; you can redistribute it and/or modify
                it under the terms of the SMT License Agreement as provided
                in the file LICENSE.TXT.  This software is distributed in
                the hope that it will be useful, but without any warranty.
 ------------------------------------------------------------------</Prolog>-*/

package com.imatix.ServletRunner;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.PrintWriter;
import java.text.DateFormat;
import java.util.Date;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;
import javax.servlet.ServletOutputStream;
import javax.servlet.http.*;

/**
    iMatix Servlet Runner implementation of 
    javax.servlet.http.HttpServletResponse.
 */
public class HttpServletResponseImpl implements HttpServletResponse
{
    private static final String DEFAULT_CONTENT_TYPE = "text/plain";

    private Vector cookieJar         = null;     // Vector of Cookies
    private Hashtable headerFields   = null;     // Hashtable of headers
    private int statusCode           = 0;        // 0 is special = not set
    private String statusMessage     = null;
    private ByteArrayOutputStream os = null;
    private PrintWriter pw           = null;
    private ServletOutputStream sos  = null;

    // Constructor -- just initialises some of the fields
    //
    HttpServletResponseImpl()
    {
       cookieJar    = new Vector(5);
       headerFields = new Hashtable();
    }

    // The following functions are in the HttpServletResponse interface

    /**
        Add cookie to the response.  It can be called multiple times to
	set more than one cookie.  The cookies are stored in a Vector
	so the number of cookes is limited only by available memory.
     */
    public void addCookie(Cookie cookie)
    {
        cookieJar.addElement(cookie); 
    }

    /**
        Returns true if the response message header has a field with
	the specified name.  Note: this method knows only about headers
	which are explicitly added using the provided methods.
     */
    public boolean containsHeader(String name)
    {
        return headerFields.containsKey(name);
    }

    /**
        Sets the status return code (and message) to return for
	this response.  Only the last return code/message specified will
	actually be used.  The message will be used as the BODY of a
	default HTML page.
     */
    public void setStatus(int sc, String sm)
    {
        statusCode    = sc;
	statusMessage = 
	      "<HTML>\n<HEAD><TITLE>Status " + sc + "</TITLE></HEAD>\n" +
	      "<BODY>" + sm + "<BODY>\n</HTML>";
    }

    /**
        Sets the status return code to return for this response.
	This method should be used when there is no error message, eg
	with SC_MOVED_TEMPORARILY.
     */
    public void setStatus(int sc)
    {
        statusCode    = sc;
	statusMessage = null;
    }

    /**
        Add a field to the response header with the given name and
	value.  If the field has already been set the new value will
	override the previous value.
     */
    public void setHeader(String name, String value)
    {
        headerFields.put(name, value);
    }

    /**
        Add an integer value to the response header under the given
	name.  If the field has already been set the new value will
	override the previous value.  The integer is turned into a
	String prior to being put into the table of header fields.
     */
    public void setIntHeader(String name, int value)
    {
        setHeader(name, Integer.toString(value));
    }

    /**
        Add a date value to the response header under the given name.
	The date is specified as milliseconds since the epoch.  If the
	field has already been set the new value will override the
	previous value.  The date is formatted, using the default
	formatting for the default Locale, prior to being put into the
	header fields.
     */
    public void setDateHeader(String name, long date)
    {
        setHeader(name, 
	          DateFormat.getDateInstance().format(new Date(date)));
    }

    /**
        Send an error response to the client using the specified status
	code and description.  The message is sent as the body of an
	HTML page, returned to the user to describe the problem.
     */
    public void sendError(int sc, String msg) throws IOException
    {
        statusCode = sc; 
	statusMessage = 
	      "<HTML>\n<HEAD><TITLE>Error " + sc + "</TITLE></HEAD>\n" + 
	      "<BODY>" + msg + "</BODY>\n</HTML>";
    }

    /**
        Send an error response to the client with the specified status
	code and a default description.  The default description is
	currently the same for all error codes.
     */
    //	XXX: Is there a way that we can easily translate the messages
    //	into a text string?
    public void sendError(int sc) throws IOException
    {
        statusCode = sc; 
	statusMessage = 
	      "<HTML>\n<HEAD><TITLE>Error " + sc + "</TITLE></HEAD>\n" +
	      "<BODY>An error of type " + sc + " has occured</BODY>\n" +
	      "</HTML>";
    }

    /**
        Send a (temporary) redirect to the client using the specified
	new location URL.  The URL must be an absolute URL.  The URL
	should have been pased through encodeRedirectURL prior to being
	used here in order to handle potential session IDs.
     */
    public void sendRedirect(String location) throws IOException
    {
        setHeader("Location", location);
	setStatus(HttpServletResponse.SC_MOVED_TEMPORARILY,
	          "The resource you requested has been moved temporarily.  " +
		  "Your browser should automatically take you to the new " +
		  "location.  If not follow the link below to vist the page " +
		  "<BR>\n<A HREF=\"" + location + "\">" + location + "</A>");
    }

    /** 
        Encodes the URL to include the session ID in it if required,
	otherwise returns it unchanged.  All URLs sent out by the
	servlet should be passed through this function to enable URL
	rewriting to be used to track sessions.

	XXX: Not implemented yet (just returns the URL unchanged).  Only
	cookies are used for session management at present.
     */
    public String encodeUrl(String url)
    {
        return url;
    }

    /**
        Encodes the specified URL for use in a sendRedirect method,if
	encoding is required, otherwise returns the URL unchanged.  All
	URLs sent out by the servlet through sendRedirect should be
	passed through this function to enable URL rewriting to be used
	to track sessions.

	XXX: Not implemented yet (just returns URL unchanged)  Only
	cookies are used for session management at present.
     */
    public String encodeRedirectUrl(String url)
    {
        return url;
    }

    // The following methods are in the ServletResponse interface which
    // is extended by the HttpServletResponse interface.

    /** 
        Set the content length of the response.
	This value is saved, and will be returned to the web server, but 
	is not validated at present.  If it is incorrect then some of
	the output may be truncated, or the browser may "hang" expecting
	more output.
     */
    public void setContentLength(int len)
    {
        setIntHeader("Content-Length", len);
    }

    /** 
        Sets the content type for the response.  If a text response will
	be written, this method must be called before starting to write
	the text response.  The method must be called only once
	(later calls will be ignored).

	As implemented, can be called only once but when it is called
	is not checked, and later calls will still result in it being
	in the header.
     */
    public synchronized void setContentType(String type)
    {
        if (! containsHeader("Content-Type"))
	    setHeader("Content-Type", type);
    }

    /**
        Get an output stream for writing binary data.

        This is implemented by taking a ByteArrayOutputStream and
	passing it to our implementation of the ServletOutputStream.

	Throws IllegalStateException if getWriter has been called on
	this same request, but tries to cope with being called twice.
     */
    public ServletOutputStream getOutputStream() throws IOException
    {
        if (sos != null)
	    return sos;                 // return existing output stream

	if (os != null)
	    throw new IllegalStateException();   // Already have output stream
	                                         // but not Servlet OS.

        // New request -- open the appropriate things and return them
        os = new ByteArrayOutputStream ();
	sos = new ServletOutputStreamImpl (os);
	return sos;
    }

    /**
        Get a print writer for writing formatted text response.
	The content type of the output must be set first, and a charset=
	property may be added to specify the character encoding.

	This is implemented by taking a ByteArrayOutputStream, and
	wrapping a PrintWriter around it.  We don't bother with
	buffering and the like because it's all going into a buffer
	anyway.

	Throws IllegalStateException if getOutputStream has been called
	on this same request, but tries to cope with being called twice.
     */
    public PrintWriter getWriter() throws IOException
    {
        if (pw != null)
	    return pw;          // return existing Writer

	if (os != null)
            throw new IllegalStateException();   // Already have output stream,
	                                         // but not a writer

        // New request -- open the appropriate things, and return them
        os = new ByteArrayOutputStream();
	pw = new PrintWriter(os);
	return pw;
    }

    /**
        Get the character set encoding for this MIME body.  This is
	either the one in the assigned content type, or one which
	the client understands.  The default is text/plain.

	XXX: At the moment we just return text/plain; but I'm not sure
	if this is the correct approach.  Maybe we should be figuring
	out how the characters are encoded (ie, character set) and
	returning that?
     */
    public synchronized String getCharacterEncoding()
    {
        if (headerFields.containsKey("Content-Type"))
	    return (String)headerFields.get("Content-Type");
	else
	    return DEFAULT_CONTENT_TYPE;
    }

    // Implementation specific methods ------------------------------------
    //
    // These methods are all accessible to the package only, and are
    // used to get at internal data for passing to other classes/methods.

    /**
        Get data output by the servlet.  This assembles all the bits
	output by the Servlet into a stream of bytes that can be
	returned to the web server.  If an error code has been set, or a
	status code with a message, then that is used in preference to
	any output that may have been written to the output stream.
     */
    byte[] getOutputData() throws IOException
    {
        // Close open streams, so we can pick up all the data
        if (pw != null)
	{
	    pw.flush();
	    pw.close();
	    pw = null;
	}

	if (sos != null)
	{
 	    sos.flush();
	    sos.close();
	    sos = null;
	}

        if (os != null)
	{
	    os.flush();
    	    os.close();
	}

	// Now assemble the output.  Potentially we have:
	// 1.  A status code and message
	// 2.  Header fields (including cookie definitions)
	// 3.  Text (etc) output to the output stream
	//
	// The logic used is:
	// If status code _and_ message, 
	//     use just the status code associated message, and the 
	//     header fields, if any.
	// Else
	//     use status code (if any), followed by the header fields
	//     (if any), followed by the text of the message (if any).

        // Assemble any headers into a single string (buffer)
	StringBuffer headers = new StringBuffer("");
	Enumeration headerEnum = headerFields.keys();
	while (headerEnum.hasMoreElements())
	{
	   String headerName = (String)headerEnum.nextElement();
           headers.append(headerName + ": " + 
	                  (String)headerFields.get(headerName) + "\n");
	}
	Enumeration cookieEnum = cookieJar.elements();
	while (cookieEnum.hasMoreElements())
	{
	   Cookie cookie = (Cookie)cookieEnum.nextElement();
	   StringBuffer cookieHeader = new StringBuffer(
	                                   "Set-Cookie: " +
	                                   cookie.getName() + "=\"" + 
				           cookie.getValue() + "\"");

           if (cookie.getVersion() > 0)
	       cookieHeader.append("; Version=\"" + 
	                           Integer.toString(cookie.getVersion()) +"\"");

           if (cookie.getPath() != null)
	       cookieHeader.append("; Path=\"" + cookie.getPath() + "\"");

         
           if (cookie.getComment() != null)
	       cookieHeader.append("; Comment=\"" + cookie.getComment() + "\"");

	                   
           if (cookie.getDomain() != null)
	       cookieHeader.append("; Domain=\"" + cookie.getDomain() + "\"");

           if (cookie.getMaxAge() > 0)
	       cookieHeader.append("; MaxAge=\"" + 
	                           Integer.toString(cookie.getMaxAge()) + "\"");

           if (cookie.getSecure())
	       cookieHeader.append("; Secure");

           cookieHeader.append("\n");

	   headers.append(cookieHeader);
	}
	if (headers.length() > 0)
  	    headers.append("\n");        // Separate headers from body


        // Now put it all together, and send it back
	//
	// XXX: We need a way to translate HTTP values into descriptions
	//
        if (statusMessage != null)       // Message and headers only
	{
	    if (headers.length() == 0)
	        headers.append("\n");    // Separate headers from body

            String startResponse = "HTTP/1.0 " + statusCode + " SeeCode\n";
            String response = startResponse + headers + statusMessage;

	    return response.getBytes();
	}
	else                             // Headers and output body 
	{
	    String startResponse = "";
	    if (statusCode != 0 && statusCode != HttpServletResponse.SC_OK)
	    {
	        startResponse = "HTTP/1.0 " + statusCode + " SeeCode\n";

		if (headers.length() == 0)
		    headers.append("\n");// Separate headers from body
	    }

            String response = startResponse + headers + os.toString();

	    return response.getBytes();
	}
    }
}
