/*  ----------------------------------------------------------------<Prolog>-
    Name:       com.imatix.ServletRunner.LRWP_1_0_WebServerConnection
    Title:      LRWP 1.0 implementation of connection to web server
    Package:    iMatix Servlet Runner

    Written:    99/01/11  Ewen McNeill <ewen@imatix.com>
    Revised:    99/03/19

    Copyright:  Copyright (c) 1991-99 iMatix Corporation
    License:    This is free software; you can redistribute it and/or modify
                it under the terms of the SMT License Agreement as provided
                in the file LICENSE.TXT.  This software is distributed in
                the hope that it will be useful, but without any warranty.
 ------------------------------------------------------------------</Prolog>-*/

package com.imatix.ServletRunner;

import java.util.Hashtable;
import java.net.Socket;
import java.io.DataInputStream;
import java.io.DataOutputStream;

/**
    LRWP 1.0 implementation of persistent connection to web server to
    retrieve and respond to requests for servlets.
 */
public class LRWP_1_0_WebServerConnection implements WebServerConnection
{
    private static final int LRWP_lenSize = 9;

    private String  webServerHostname     = "localhost"; // Default Web Server
    private int     LRWP_Port             = 81;          // Default LRWP port
    private boolean connectionEstablished = false;
    private boolean haveConnected         = false;       // Connected in past
    private String  errorMessage          = "Looking good so far";

    private Socket  LRWP_Connection       = null;
    private DataInputStream LRWP_In       = null;
    private DataOutputStream LRWP_Out     = null;

    private byte[]  inputData             = null;
    private byte[]  CGI_environment       = null;
    private Hashtable requestContext      = null;

    /**
        Configure web server, and port to connect to when establishing
	connection.  These details are expected to remain static, while
	connections can be brought up and torn down at will.
     */
    public LRWP_1_0_WebServerConnection (String webServerHostname,
                                         int    LRWP_Port)
    {
        super();
        this.webServerHostname = webServerHostname;
	this.LRWP_Port         = LRWP_Port;
    }

    /**
        Configure webserver to use when establishing connection.  The
	port used is the default one.
     */
    public LRWP_1_0_WebServerConnection (String webServerHostname)
    {
        super();
        this.webServerHostname = webServerHostname;
    }

    /**
        Configure object to use default web server hostname and port.
     */
    public LRWP_1_0_WebServerConnection ()
    {
        super();
    }

    /**
        Establish connection to the web server to indicate willingness
	to handle connections for the given applicationName on a given
	virtualHost.  The virtualHost may be unspecified (null), which
	means that the applicationName will apply on all hosts.  The
	accessKey is unused.

	If this function retruns false, an indicative error message can
	be retrieved with getErrorMessage().
     */
    public boolean connectToWebServer (String applicationName,
                                       String virtualHost,
				       String accessKey)
    {
        final int LRWP_ResponseLen = 1024;
	errorMessage = "Looking good so far";

	// Tidy up old connection (if any)
	connectionEstablished = false;
        inputData             = null;
	CGI_environment       = null;
	requestContext        = null;
	try {
	    LRWP_Connection.close();
	}
	catch (NullPointerException e) {}
	catch (Exception e) {
	    errorMessage = "Cannot close old connection: " + e.toString();
	    return false;
	}

        // Establish new connection to web server
	try {
	    LRWP_Connection = new Socket (webServerHostname, LRWP_Port);
	    LRWP_Connection.setSoTimeout (0);
	}
	catch (Exception e) {
	    errorMessage = "Cannot connect to web server: " + e.toString();
	    return false;
	}

	// Build input and output streams to use with socket
	try {
	    LRWP_In = new DataInputStream (LRWP_Connection.getInputStream ());
	}
	catch (Exception e) {
	    errorMessage = "Cannot establish input stream: " + e.toString();
	    try {
	        LRWP_Connection.close();
	    }
	    catch (Exception ex) {}
	    return false;
	}

	try {
	    LRWP_Out = 
	         new DataOutputStream (LRWP_Connection.getOutputStream ());
	}
	catch (Exception e) {
	    errorMessage = "Cannot establish output stream: " + e.toString();
            try {
	        LRWP_In.close();
		LRWP_Connection.close();
	    }
	    catch (Exception ex) {}
	    return false;
	}

	// Now let the LRWP Server know what we want to handle.
	// The format of the message is: (LRWP 1.0)
	//     applicationName \0xFF virtualHostName \0xFF "filter" \0xFF
	//
	// We do not send a filter string, but for full compatibility do
	// send an empty filter string and end marker in its place.
	//
	int subscribeLength = 
	    applicationName.length() + 
	    (virtualHost != null ? virtualHost.length() : 0) + 3;
	byte[] subscribeRequest = new byte[subscribeLength];
        System.arraycopy (applicationName.getBytes(), 0,
	                  subscribeRequest, 0, applicationName.length ());

        subscribeRequest [applicationName.length ()] = (byte)255;
	if (virtualHost != null)
	{
	    System.arraycopy (virtualHost.getBytes(), 0,
	                      subscribeRequest,
			      (applicationName.length () + 1),
			      virtualHost.length());
	}
	subscribeRequest [applicationName.length () + 1 +
	                  (virtualHost != null ? virtualHost.length() : 0)]
                 = (byte)255;
        subscribeRequest [subscribeRequest.length - 1] = (byte)255;

        try {
	    LRWP_Out.write(subscribeRequest, 0, subscribeRequest.length);
	    LRWP_Out.flush();
	}
	catch (Exception e) {
	    errorMessage = "Cannot send LRWP subscription: " + e.toString();
            try {
	        LRWP_In.close();
		LRWP_Out.close();
		LRWP_Connection.close();
	    }
	    catch (Exception ex) {}
	}

	// Wait for acknowledgement from LRWP server
	// Under LRWP 1.0 this consists either of an error message, or
	// the word "OK".  In either case we assume we'll receive it all
	// in one bunch (ie ignoring networking issues).
	byte[] subscribeResponse = new byte [LRWP_ResponseLen];
	int    subscribeRespLen  = 0;
	try {
	    subscribeRespLen = LRWP_In.read (subscribeResponse);
	}
	catch (Exception e) {
	    errorMessage = "Could not read response: " + e.toString();
            try {
                LRWP_In.close();
                LRWP_Out.close();
                LRWP_Connection.close();
            }
            catch (Exception ex) {}
	}

	String subscribeResponseText = new String (subscribeResponse, 0,
	                                           subscribeRespLen);
	if (subscribeResponseText.startsWith ("OK"))
	{
	    connectionEstablished = true;
	    haveConnected         = true;
	    return true;
        }

	// If we got through this far, we've received some sort of error
	// message.  The connection has probably been closed on us, but
	// we'll ignore that for now.
	errorMessage = "Subscription Refused: " + subscribeResponseText;
	return false;
    }

    /**
        Wait for new requests, and accept the details of the request.
	The actual details can be retrieved with other functions, such
	as getRequestContext(), and getRequestDate(), to retrieve the
	equivilent of the CGI environment, and the posted data (if any).

	The exception NotConnectedToWebServer will be thrown if the
	connection to the web server has not been established yet, or
	the connection has been lost.
     */
    public boolean waitForRequest () throws NotConnectedToWebServer
    {
        errorMessage = "Looking good so far";

        if (! connectionEstablished)
	{
	    throw new NotConnectedToWebServer 
	              ("Never connected, or have disconnected");
	}

	if (inputData != null || CGI_environment != null)
	{
	    errorMessage = "Connection already in progress";
	    return false;
        }

	byte [] lengthBuf = new byte [LRWP_lenSize];

        // Read in a new request.
	// In LRWP 1.0 the request consists of:
	//    9 bytes of environmentLength
	//    environmentLength bytes of environment text
	//    9 bytes of postedDataLength
	//    postedDataLength bytes of posted Data
	//
	try {
            // Obtain the environment
	    //
	    LRWP_In.readFully (lengthBuf, 0, lengthBuf.length);
	    int environmentLength = Integer.parseInt (new String (lengthBuf));
	    CGI_environment = new byte [environmentLength];
	    LRWP_In.readFully (CGI_environment, 0, CGI_environment.length);

	    // Get the posted data (if any) 
	    LRWP_In.readFully (lengthBuf, 0, lengthBuf.length);
	    int postedDataLength = Integer.parseInt (new String (lengthBuf));
	    if (postedDataLength > 0)
	    {
	        inputData = new byte [postedDataLength];
		LRWP_In.readFully (inputData, 0, inputData.length);
	    }

	    return true;             // Have a new request
	}
	catch (Exception e) {
	    errorMessage = "Error receiving request: " + e.toString();
	    return false;
        }
    }

    /**
        Retrieve the context of the request.  This returns a hash table
	consisting of (at least) the elements that are available in the
	environment to a CGI program when run from a CGI 1.1 compliant
	server.

	The exception NotConnectedToWebServer will be thrown if the
	connection to the web server has not been established yet, or
	the connection has been lost.
	The exception NoActiveRequest will be thrown if there is not
	an open request at the time this function is called.
      */
    public Hashtable getRequestContext () throws NoActiveRequest
    {
        if (! connectionEstablished)
	    throw new NotConnectedToWebServer
	              ("Never connected or have disconnected");
	
        if (CGI_environment == null)
            throw new NoActiveRequest ("No request environment present"); 

    	if (requestContext == null)
	    requestContext = parseEnvironmentStrings (CGI_environment);

        return requestContext;
    }

    /**
        Retrieve the input data from the request, if any.  Returns 
	null if there is no input data to be processed.

	The exception NotConnectedToWebServer will be thrown if the
	connection to the web server has not been established yet, or
	the connection has been lost.
	The exception NoActiveRequest will be thrown if there is not
	an open request at the time this function is called.
     */
    public byte[] getRequestInput () throws NoActiveRequest
    {
        if (! connectionEstablished)
	    throw new NotConnectedToWebServer 
	              ("Never connected or have disconnected");
        if (CGI_environment == null)
	    throw new NoActiveRequest ("No request active at present");

        return inputData;
    }

    /** 
        Send response to the web server.  The response is provided as
	an array of bytes to send to the webserver.

	The exception NotConnectedToWebServer will be thrown if the
	connection to the web server has not been established yet, or
	the connection has been lost.
	The exception NoActiveRequest will be thrown if there is not
	an open request at the time this function is called.
     */
    public boolean sendResponse (byte[] response, int startAt, int endAt)
    {
        if (! connectionEstablished)
	    throw new NotConnectedToWebServer
	              ("Never connected, or have disconnected");
        if (CGI_environment == null)
	    throw new NoActiveRequest ("No request in progress to respond to");

	// Calculate the length value to send across the LRWP
	// connection.  This has to be 0 padded to the right, which is a
	// touch tricky in Java.
	//
	int responseLength = endAt - startAt + 1;
	String lengthStr = String.valueOf(responseLength);

	byte[] sendLength = new byte [LRWP_lenSize];
	for (int i = 0; i < sendLength.length; i++)
	    sendLength [i] = (byte)'0';

        System.arraycopy(lengthStr.getBytes(), 0,
	                 sendLength, (LRWP_lenSize - lengthStr.length()),
			 lengthStr.getBytes ().length);

        // Now send out the response, in LRWP format.  That is the 
	// length first, and then the actual data.
        try {
	    LRWP_Out.write(sendLength, 0, sendLength.length);
	    LRWP_Out.write(response, startAt, responseLength);

	    LRWP_Out.flush();
	}
	catch (Exception e)
	{
	    errorMessage = "SendResponse error: " + e.toString();
	    return false;
	}

	// Now clear the request status, ready for a new response
	CGI_environment = null;
	inputData       = null;
	requestContext  = null;

	return true;
    }
    public boolean sendResponse (byte[] response)
    {
        return sendResponse (response, 0, (response.length - 1));
    }
    public boolean sendResponse (byte[] response, int startAt)
    {
        return sendResponse (response, startAt, (response.length - 1));
    }

    /**
        Retrieve error message (if any).  The error message is useful 
	only if an error has occured, otherwise it contains an
	innocent string.
     */
    public String getErrorMessage()
    {
        return errorMessage;
    }

    /** 
        Disconnect from the web server.  This can be done while a
	request is in progress, but will result in that request being
	abandoned.  If there is no active connection to the web server,
	this method will be a noop.
     */
    public void disconnectFromWebServer()
    {
        if (! connectionEstablished)
	    return;

        try {
	    LRWP_In.close();
	    LRWP_Out.close();
	}
	catch (Exception e) { }
        try {
	    LRWP_Connection.close();
	}
	catch (Exception e) { }

	connectionEstablished = false;
    }

    /**
        Parse environment strings.  The environment is passed over 
	as a single long string, which contains pairs of name=value
	separated by NULs.  In parsing them, we look for the first
	"=", then look for the first NUL.  The bit up to the "="
	is the name, and the bit from there to the NUL is the value.
	This is repeated until the end of the string.
     */
    private Hashtable parseEnvironmentStrings (byte [] environmentData)
    {
        if (environmentData == null)
	    return null;

        // Guestimate the size of the hash table that will be required.
	// 60 is a random value guesstimated to be about the average
	// length of an entry in the environment.  This is done for
	// efficiency (the hash table will grow as required).
	//
	int hashTableSize = (int)(environmentData.length / 60);
	Hashtable env = new Hashtable (hashTableSize);

	// Now scan through the bytes in the environment, looking for
	// pairs.
        int    nameStart  = 0;
	int    nameEnd    = 0;
	int    valueStart = 0; 
	int    valueEnd   = 0;

	final byte nameValSep = (byte)'=';
	final byte pairSep    = (byte)0;

        while (nameStart < environmentData.length)
	{
            while (nameEnd < environmentData.length &&
	           environmentData [nameEnd] != nameValSep &&
		   environmentData [nameEnd] != pairSep)
                ++nameEnd;

	    if (nameEnd < environmentData.length &&
	        environmentData [nameEnd] == nameValSep)
	    {
	        valueStart = nameEnd + 1;
		valueEnd   = valueStart;
		--nameEnd;
	    }
            else
                break;

	    while (valueEnd < environmentData.length &&
	           environmentData [valueEnd] != pairSep)
                ++valueEnd;

	    if (valueEnd < environmentData.length &&
	        environmentData [valueEnd] == pairSep)
	        --valueEnd;
            else
	        if (valueEnd >= environmentData.length)
		    valueEnd = environmentData.length - 1;

	    // Now: name is in range nameStart->nameEnd
	    // and  value is in range valueStart->valueEnd
	    //
	    // Build up the hash table entry, we let String to the
	    // chopping out of the array of bytes, since it is good at it.
	    //
	    env.put (new String (environmentData, 
	                         nameStart, 
				 (nameEnd-nameStart+1)),
		     new String (environmentData,
		                 valueStart,
				 (valueEnd-valueStart+1)));

	    nameStart = valueEnd + 2;      // Skip the NUL
	    nameEnd = valueStart = valueEnd = nameStart;
	}

	return env;
    }

    /**
        Test harness.
      */
    public static void main (String args[])
    {
        String appName = "lrwptest";
	String webServer = "localhost";
	String virtualHost = null;
	String accessKey = null;

        // Look for arguments that are around; use defaults for any that
	// are not present.
	//
	if (args.length > 0)
	    appName = args[0];
	if (args.length > 1)
	    webServer = args[1];
	if (args.length > 2)
	    virtualHost = args[2];
	if (args.length > 3)
	    accessKey   = args[3];


        // Now establish a connection to the web server
	//
        WebServerConnection conn = 
	    new LRWP_1_0_WebServerConnection (webServer);
	if (! conn.connectToWebServer(appName, virtualHost, accessKey))
	{
	    System.out.println("Could not connect to web server");
	    System.out.println("Error was: " + conn.getErrorMessage ());
	}

	int requestNo = 0;

        for (requestNo = 1; requestNo < 10; requestNo++)
	{
	    if (conn.waitForRequest())
	    {
	        System.out.println("New request...");
		System.out.println("Env: " + conn.getRequestContext ());

		System.out.println("Posted Data: " + conn.getRequestInput ());

                // Send back a quick test resposne 
		String response = 
		    "Content-type: text/plain\n\nResponse " + requestNo;

		if (! conn.sendResponse( response.getBytes () )) {
		    System.out.println ("Error sending response: " + 
		                         conn.getErrorMessage ());
		    break;
		}
	    }
	    else
	    {
	        System.out.println ("Error accepting request: " + 
		                     conn.getErrorMessage ());
                break;
	    }
	}

	conn.disconnectFromWebServer();
    }
}
