/*  ----------------------------------------------------------------<Prolog>-
    Name:       com.imatix.ServletRunner.ServletRunInfo
    Title:      Information about running Servlets
    Package:    iMatix Servlet Runner

    Written:    99/03/19  Ewen McNeill <ewen@imatix.com>
    Revised:    99/03/19

    Copyright:  Copyright (c) 1991-99 iMatix Corporation
    License:    This is free software; you can redistribute it and/or modify
                it under the terms of the SMT License Agreement as provided
                in the file LICENSE.TXT.  This software is distributed in
                the hope that it will be useful, but without any warranty.
 ------------------------------------------------------------------</Prolog>-*/

package com.imatix.ServletRunner;

import javax.servlet.http.HttpServlet;
import javax.servlet.ServletConfig;
import javax.servlet.ServletException;
import javax.servlet.SingleThreadModel;

/**
    This class maintains information about a servlet that is running,
    including the minimum and maximum number of free Connections, access
    statistics, and a reference to the servlet itself (if instantiated).
    The last access time is also recorded, to potentially allow removal
    of unused servlets.
 */
public class ServletRunInfo
{
    private final boolean debug = false;   // For Debugging load balancing

    private int minFreeConnections;
    private int maxFreeConnections;
    private int curConnections;
    private int busyConnections;

    private long numAccess;
    private long lastAccess;              /* System millisecond timer */

    private HttpServlet    servlet;
    private Class          servletClass;
    private ServletConfig  config;

    protected ServletRunInfo(int min, int max,
                             Class          theServletClass,
			     ServletConfig  theServletConfig)
	throws IllegalArgumentException
    {
      if (min < 0 || min >= max)
          throw new IllegalArgumentException 
	            ("Connections must be: 0 < min < max");
      if (theServletClass == null)
          throw new IllegalArgumentException
	            ("Must supply servlet class");
      if (theServletConfig == null)
          throw new IllegalArgumentException
	            ("Must supply configuration for servlet");

      minFreeConnections  = min;
      maxFreeConnections  = max;
      curConnections      = 0;
      busyConnections     = 0;
      numAccess           = 0;
      lastAccess          = 0;
      servlet             = null;
     
      servletClass        = theServletClass;
      config              = theServletConfig;
    }

    protected int getMinFreeConnections()      { return minFreeConnections; }
    protected int getMaxFreeConnections()      { return maxFreeConnections; }

    protected synchronized int getCurConnections() { return curConnections; }
    protected synchronized int getBusyConnections(){ return busyConnections; }

    protected synchronized void busyConnection()   
    { ++busyConnections; 
      if (debug)
          System.err.println("busyConnection: " + 
	                     "Currently busy: " + busyConnections + "/" +
	                     curConnections);
    }
    protected synchronized void freeConnection()   
    { --busyConnections;
      if (debug)
          System.err.println("freeConnection: " +
	                     "Currently busy: " + busyConnections + "/" +
	                     curConnections);
    }

    protected synchronized void newConnection()    
    { ++curConnections;
      if (debug)
          System.err.println("newConnection: " +
	                     "Currently busy: " + busyConnections + "/" +
	                     curConnections);
    }
    protected synchronized void removeConnection() 
    { --curConnections; 
      if (debug)
          System.err.println("removeConnection: " +
	                     "Currently busy: " + busyConnections + "/" +
	                     curConnections);
    }

    protected synchronized boolean moreFreeThanMax() 
    { boolean morefree = 
              (curConnections - busyConnections) > maxFreeConnections; 
      if (debug)
          System.err.println("moreFreeThanMax: " + 
	                     (morefree ? "true" : "false"));
      return morefree;
    }

    protected synchronized boolean fewerFreeThanMin()
    { boolean fewerfree =
          (curConnections - busyConnections) < minFreeConnections; 
      if (debug)
          System.err.println("fewerFreeThanMin: " + 
	                     (fewerfree ? "true" : "false"));
      return fewerfree;
    }

    /**
        Get instance of servlet, creating one if required.
      */
    protected synchronized HttpServlet getServlet()
              throws InstantiationException, IllegalAccessException,
	             ServletException
    {
        lastAccess = System.currentTimeMillis();
	++numAccess;

        if (servlet != null)
	    return servlet;

        // Nothing in stock; need to manufacturer it ourselves
        servlet = (HttpServlet)servletClass.newInstance();
        servlet.init(config);

        // Now we have the class, figure out if single thread model applies
	if (servlet instanceof SingleThreadModel) 
	{
	    minFreeConnections = 0;
	    maxFreeConnections = 9999;
	}

        return servlet;
    }

    /** 
        Get instance of servlet (if any).
      */
    protected synchronized HttpServlet getExistingServlet()
    {   return servlet; }
}
