/*  ----------------------------------------------------------------<Prolog>-
    Name:       com.imatix.ServletRunner.ServletRunner
    Title:      Main startup code for Servlet Runner
    Package:    iMatix Servlet Runner

    Written:    99/01/11  Ewen McNeill <ewen@imatix.com>
    Revised:    99/03/19

    Copyright:  Copyright (c) 1991-99 iMatix Corporation
    License:    This is free software; you can redistribute it and/or modify
                it under the terms of the SMT License Agreement as provided
                in the file LICENSE.TXT.  This software is distributed in
                the hope that it will be useful, but without any warranty.
 ------------------------------------------------------------------</Prolog>-*/

package com.imatix.ServletRunner;

import java.io.BufferedInputStream;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.PrintWriter;
import java.text.DateFormat;
import java.util.Date;
import java.util.Calendar;
import java.util.Dictionary;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Locale;
import java.util.Properties;
import java.util.TimeZone;
import javax.servlet.http.*;
import javax.servlet.*;

/**
    This class is the main class of the iMatix Servlet Runner.  It
    is responsible for loading the property files, and the servlets, and
    establishing their connections to the web server.  It operates with
    one thread per servlet at present (implicitly treating all servlets
    as single threaded).  A later update will handle multithreaded
    servlets and dynamically establish (and tear down) LRWP connections.
 */
public class ServletRunner implements Runnable, ServletController
{
    // Class Constants
    static final String MYNAME   = "iMatix Servlet Runner";

    static final long REQUEST_SLEEP = 250;   // Debug sleep time (ms)

    // Class-wide variables, and their defaults.  These are primarily
    // overall configuration information.
    static String servletPrefix  = "servlet/";
    static String webServer      = "localhost";
    static String virtualHost    = null;
    static int    LRWPportNo     = 81;
    static String propertiesFile = "servlet.properties";
    static String logFileName    = "servlet.log";

    // For multithreaded servlets, number of free connections to maintain
    //
    // NOTE: It is recommended that maxFreeConnections >> minFreeConnections
    // (that is, much greater than), so that the free connections can
    // ramp up fairly quickly, but closes back down again slowly.
    // Closing connections may cause requests to the servlet to fail
    // with an internal error.
    //
    static int    minFreeConnections = 1;
    static int    maxFreeConnections = 10;

    // A hash of servlet names to servlet instances, a log file, and the
    // servlet Properties (initialised in main).
    //
    // The servlet name->instances hash returns ServletRunInfo classes
    // which record an instance of the servlet, and min/max/access info.
    // 
    static Hashtable servlets    = null;
    static PrintWriter log       = null;
    static Properties servletProperties = null;

    // Set to true for debugging messages
    static final boolean DEBUG = false;

    /**
        Parse command line arguments, read in servlet properties, and
	load servlets into memory, establishing their connections to
	the web server.
     */
    public static void main (String args[])
    {
	boolean showUsage = false;

        // XXX: Is there a tidier way to parse arguments?  A getopt() clone?
	//
	for (int argNo = 0; argNo < args.length; argNo++)
	{
	    if      (args[argNo].equals("-w"))    // Web Server Host
	        if (argNo + 1 < args.length)
		    webServer = args[++argNo];
		else
		    showUsage = true;
            else if (args[argNo].equals("-p"))   // LRWP port
	        if (argNo + 1 < args.length)
		    try {
		        LRWPportNo = Integer.parseInt(args[++argNo]);
		    }
		    catch (NumberFormatException e) {
		        showUsage = true;
		    }
		else 
		    showUsage = true;
            else if (args[argNo].equals("-s"))   // Servlet Properties File
                if (argNo + 1 < args.length)
		    propertiesFile = args[++argNo];
		else
		    showUsage = true;
            else if (args[argNo].equals("-u"))   // servlet URL prefix
	        if (argNo + 1 < args.length)
		    servletPrefix = args[++argNo];
		else
		    showUsage = true;
            else if (args[argNo].equals("-v"))   // virtual host (default)
	        if (argNo + 1 < args.length)
		    virtualHost = args[++argNo];
		else
		    showUsage = true;
            else if (args[argNo].equals("-l"))   // log file name
	        if (argNo + 1 < args.length)
		    logFileName = args[++argNo];
		else
		    showUsage = true;
            else if (args[argNo].equals("-m"))   // min, max free instances
	        if (argNo + 2 < args.length)
		    try {
		        minFreeConnections = Integer.parseInt(args[++argNo]);
		        maxFreeConnections = Integer.parseInt(args[++argNo]);
		    }
		    catch (NumberFormatException e) {
		        showUsage = true;
		    }
                else
		    showUsage = true;
            else
	        showUsage = true;
	}

	if (showUsage)
	{
	    System.out.println("iMatix Servlet Runner\n");
	    System.out.println("Usage:");
	    System.out.println("  -w web_server    -- hostname of web server");
	    System.out.println("  -p port_no       -- LRWP port of web server");
	    System.out.println("  -s servlet_props -- Servlet Properties file");
	    System.out.println("  -u url_prefix    -- Servlet URL prefix");
	    System.out.println("  -v virtual_host  -- virtual host of server");
	    System.out.println("  -l log_file      -- name of servlet log");
	    System.out.println("  -m min max       -- free LRWP connections");
	    System.exit(1);
	}

	// Load properties file, merging in with a copy of the system 
	// properties.
	//
	servletProperties = new Properties(System.getProperties());
	try {
	    if (DEBUG)
  	        System.err.println("Trying to load servlet properties from " + 
	                           propertiesFile);
	    servletProperties.load(
	           new BufferedInputStream(
		       new FileInputStream(propertiesFile)));

            if (DEBUG)
                System.err.println("Current properties are:\n" +
	                            servletProperties);
	}
	catch (IOException e)
	{
	    System.out.println("Unable to load servlet properties from " +
	                       "file: " + propertiesFile + " -- giving up");
            System.exit(1);
	}

	// Open up log file, and announce ourselves.
	try {
	    if (DEBUG)
	        System.err.println("Opening log on: " + logFileName);

	    log = new PrintWriter(
		      new FileOutputStream(logFileName, true), // append
	              true);                                   // autoflush
	    logMsg(MYNAME + " -- started");
	}
	catch (IOException e)
	{
	    System.out.println("Unable to open log file: " + logFileName);
	    System.exit(1);
	}

	// Build a hash table for the servlets
	servlets = new Hashtable();

	// Now, for each servlet listed in the servlet properties, start
	// a thread up to handle that servlet.
	//
	// We look for keys of the form servlet.NAME.code, where NAME
	// will be the name of the servlet.  (This form is not ideal, 
	// but is the way that the JSDK 2.0 supplied servlet runner
	// works so we copy it to ease transition.)
	//
	Enumeration servletEnum = servletProperties.keys();
	while (servletEnum.hasMoreElements())
	{
	    String keyName = (String)servletEnum.nextElement();
            if (keyName.startsWith("servlet.") &&
	        keyName.endsWith  (".code"))
            {
	        int startIndex = "servlet.".length();
		int endOffset  = ".code".length();

		String servletName = 
		    keyName.substring(startIndex, keyName.length()-endOffset);

                Runnable me = new ServletRunner(servletName);
		Thread   th = new Thread(me);
		th.start();
	    }
	}

	// This thread ends, but java continues running until the other
	// threads exit.
    }

    // Instance variables
    String servletName = null;

    /**
        Constructor: builds an object to handle the named servlet,
	running in a thread of its own.
     */
    private ServletRunner(String name)
    {
        servletName = name;
    }

    // Methods required by the Runnable interface

    /**
        Handle servlet in a thread of its own.  This builds a connection
	to the web server for the servlet (using LRWP), and waits for
	requests.  Upon receiving a request it starts up the servlet (if
	required) and passes the request to the servlet, and then
	returns the servlets response to the web server.  This
	continues until the program is stopped.
     */
    public void run()
    {
        // Validate servlet details (need to be able to load class in 
	// order to do anything; no point in even connecting to web
	// server if we cannot load it).
	String servletClassName = null;
	String servletAccessKey = null;
	Class  servletClass     = null;
	try {
  	    servletClassName = (String)
	           servletProperties.get("servlet." + servletName + ".code");
            servletClass = Class.forName(servletClassName);
	}
	catch (ClassNotFoundException e) {}
	catch (NullPointerException e)   {}

	if (servletClass == null)
	{
	    logMsg(servletName, "servlet class not found " + 
	                  (servletClassName != null 
			          ? "(" + servletClassName + ")" 
				  : "(no class name)"));
            return;
	}

	try {
  	    servletAccessKey = (String)
	           servletProperties.get("servlet." + servletName + ".key");
	}
	catch (NullPointerException e)   {}

        // We have two means of connecting to the web server:
	// LRWP1 -- supported by all versions of Xitami 2.x
	// LRWP2 -- supported by Xitami >= 2.5
	//
	// LRWP2 allows password authentication which may be required
	// for access to the web server.  Hence we use LRWP if we have 
	// a password use, otherwise we use LRWP 1.

        logMsg(servletName, "Connecting to webserver with " +
	                    (servletAccessKey == null ? "LRWP1" : "LRWP2"));
        WebServerConnection conn = null;
        if (servletAccessKey == null)
	    conn = new LRWP_1_0_WebServerConnection (webServer, LRWPportNo);
        else
	    conn = new LRWP_2_0_WebServerConnection (webServer, LRWPportNo);

	if (! conn.connectToWebServer(servletPrefix + servletName, 
	                              virtualHost,
				      servletAccessKey))
	{
	    logMsg(servletName, "connection to webserver failed: " + 
	                        conn.getErrorMessage());
	    return;
	}

	// servlet Session Context -- initially empty
	// As the servlet requests sessions this will be filled up.
	//
	HttpSessionContextImpl servletSessionContext = 
	           new HttpSessionContextImpl();

	// At this point we are connected to the web server.  We
	// register ourselves in the table of instances (creating an
	// entry for the table if required).

	ServletRunInfo runinfo = null;
	synchronized (servlets) {
	    runinfo = (ServletRunInfo)servlets.get(servletName);

	    if (runinfo != null) 
	        runinfo.newConnection();
	    else 
	    {
	        // Need to create an entry because there is no existing
		// entry.
	        ServletContext sc = new ServletContextImpl(this, log);
 	        ServletConfig  config = new ServletConfigImpl(sc,
	    	          buildInitArgs(
			       (String)servletProperties.get(
				    "servlet."+servletName+".initArgs")));

                // NOTE: min/max may be overridden later if the servlet
		// is marked to be single threaded.
		try {
		    runinfo = new ServletRunInfo(minFreeConnections, 
						 maxFreeConnections,
						 servletClass, config);
                }
		catch (IllegalArgumentException e) {
		    logMsg(servletName, "Could not register servlet: " + e);
 	            conn.disconnectFromWebServer();
		    return;
		}
                runinfo.newConnection();

                servlets.put(servletName, runinfo);
	    }
	}
	
	// Wait for requests, and then process them.  Each time around
	// we obtain a refrence to the servlet, and then use that to run
	// the request and return the result.
	//
        while (true)
	{
	    if (conn.waitForRequest())
	    {
                logMsg(servletName, "request received");

                HttpServletResponseImpl resp = 
		    new HttpServletResponseImpl ();

		HttpServletRequestImpl req = 
		    new HttpServletRequestImpl (conn.getRequestContext (),
		                                conn.getRequestInput (),
						servletSessionContext,
						resp);

                // Get the servlet if it exists, otherwise start a new
		// one.  This needs to be synchronised to avoid
		// potential race later.
		HttpServlet s = null;
                synchronized (servlets)
		{
		    runinfo = (ServletRunInfo)servlets.get(servletName);

		    Exception   ex = null;
                    try {
                        s = runinfo.getServlet();
                    }
		    catch (NullPointerException   e) { s = null; ex = e; }
		    catch (InstantiationException e) { s = null; ex = e; }
		    catch (IllegalAccessException e) { s = null; ex = e; }
		    catch (ServletException       e) { s = null; ex = e; }

	            if (s == null)
 	            {
	               logMsg(servletName, "Couldn't get Server Instance: " + 
		                           ex);
 	               conn.disconnectFromWebServer();
		       return;
	            }

		    // Record this instance as being busy, and check if
		    // we should be starting another one
		    runinfo.busyConnection();

		    if (runinfo.fewerFreeThanMin())
		    {
		        logMsg(servletName, "Creating new connection");
			Runnable me = new ServletRunner(servletName);
			Thread   th = new Thread(me);
			th.start();
		    }
		} // synchronized (servlets)

		// At this point we have our servlet, it is ready to go,
		// and we have the request and response all prepared.
		//
		// So we run the request, and send the response back.
		//
                try {
		    logMsg(servletName, "running request");
	            s.service(req, resp);	

		    byte[] response = null;

		    try {
			response = resp.getOutputData();
		    }
		    catch (IOException e) { response = null; }

                    if (response == null)
		        response = new String(
			    "Content-type: text/plain\n\n" +
			    "Internal error running servlet").getBytes();

                    // DEBUGGING
		    if (REQUEST_SLEEP > 0)
		    {
		        try {
			    Thread.sleep(REQUEST_SLEEP);
			}
			catch (Exception e) {}
		    }

		    if (! conn.sendResponse( response)) {
		        logMsg(servletName, "Error sending servlet response " + 
		                            conn.getErrorMessage ());
		    }

		    // Free up this connection, we're ready for another
		    // request.  And check if we should be exiting
		    // because there are too many around.  (There's a
		    // race in the free connections, but we ignore that
		    // for the moment)
		    runinfo.freeConnection();
                    if (runinfo.moreFreeThanMax() &&
		        runinfo.getCurConnections() > 1)
			break;
		}
		catch (Exception e)
		{
		    logMsg(servletName, "Error running servlet: " +
		                        e.toString());

                    // If we got a runtime exception stop trying to run
		    // this servlet, it's probably buggy, and running it
		    // again will just cause the same problem again.
		    //
                    if (e instanceof RuntimeException)
		        break;
		}
	    }
	    else                          // Wait for request failed
	    {
	       logMsg(servletName, "no more requests: " +
	                            conn.getErrorMessage());

               break;              // If could not get request, give up
	    }
	}

	// Okay, we're done.  Remove this instance out of the servlet
	// information table.  If we're the last one left, turn out the
	// lights (eg, shut down the servlet, remove the entry)
	synchronized (servlets) {
	    runinfo.removeConnection();
	    if (runinfo.getCurConnections() == 0) 
	    {
	        servlets.remove(servletName);
		HttpServlet s = runinfo.getExistingServlet();
	        if (s != null)
                    s.destroy();
	    }
	}

	if (conn != null)
	    conn.disconnectFromWebServer();
    }

    // Methods required by the Servlet Controller interface
    /**
        Returns a reference to the internal directory of loaded 
        Servlets, which can be used to determine which other Servlets
        are running.  Contacting those servlets directly is NOT 
        recommended.  Access to the table should be synchronized against
        the returned table.
     */
    public Dictionary getServletsTable()
    {
        return servlets;
    }    

    /** 
        Translate a given path from a URI/virtual path into a path
        on the disk.
 
        XXX: not implemented yet.  Not quite clear how we should do
	the translation here, other than prepending the webpages root.
     */
    public String translatePath(String virtualPath)
    {
        return virtualPath;
    }

    /**
        Determine the mime type to be used with the file (if possible).
        Returns null if the mime type cannot be determined.

        XXX: not implemented yet.  We'll probably need to hard code
	these values.
     */
    public String determineMimeType(String fileName)
    {
        return null;
    }

    /**
        Gets the type of server in use.  This returns the static
        identification of the servlet runner.
     */
    public String getServerInfo()
    {
        return MYNAME;
    }

    /** 
        Write an entry into the log file, with the servlet name
        and time and date stamp.
     */
    public void servletLog(String message)
    {
        logMsg(servletName, message);
    }

    // Static utility methods
    /**
        Write a message to the logfile, with a time/date stamp.
     */
    // XXX: The overhead on this is fairly high.  DateFormat cannot be
    // used from two threads at once (internal race), which is why we
    // build a new one.  But since we're synchronised; maybe restructure.
    //
    // Also the formats of the dates/times is UGLY.  But the choice
    // appears to be that, or roll our own :-(
    //
    private synchronized static void logMsg(String message)
    {
	Calendar cal = 
	    Calendar.getInstance(TimeZone.getDefault(), Locale.getDefault());
	DateFormat formatter = 
	    DateFormat.getDateTimeInstance(DateFormat.MEDIUM,
	                                   DateFormat.MEDIUM,
					   Locale.getDefault());
        formatter.setCalendar(cal);
        Date now = new Date();
        log.println(formatter.format(now) + " " + message);
    }

    /**
        Write a message to the logfile for a specific servlet, recording
	the name of the servlet and a time/date stamp.
     */
    private static void logMsg(String servletName, String message)
    {
        logMsg(servletName + ": " + message);
    }

    /**
        Decode the initial arguments string (if any) to the servlet
	into a hash table, splitting on commas and equals.
     */
    private static Hashtable buildInitArgs(String servletInitArgs)
    {
        Hashtable initArgs = new Hashtable();

	if (servletInitArgs == null)
	    return initArgs;           // No entries

        // Parse string of the form:
	//    name=value,name=value,name=value,....,name=value
	// into a hash.
	//
	int nameStart  = 0;
	int nameEnd    = 0;
	int valueStart = 0;
	int valueEnd   = 0;

        final int nameValSep = (int)'=';
	final int pairSep    = (int)',';

	while (nameStart < servletInitArgs.length())
	{
	    nameEnd = servletInitArgs.indexOf(nameValSep, nameStart);
	    if (nameEnd < 0)
	        break;

            valueStart = nameEnd + 1;
	    if (valueStart > servletInitArgs.length())
	        break;

            valueEnd = servletInitArgs.indexOf(pairSep, valueStart);
	    if (valueEnd < 0)
	        valueEnd = servletInitArgs.length();
	
	    //System.err.println("Extracting name from " + nameStart +  
	    //                   " to " + nameEnd);
	    //System.err.println("Extracting value from " + valueStart +  
	    //                   " to " + valueEnd);
	    String argName = servletInitArgs.substring(nameStart, nameEnd);
	    String argValue = servletInitArgs.substring(valueStart, valueEnd);

	    initArgs.put(argName, argValue);

	    nameStart = valueEnd + 1;
	    valueStart = valueEnd = nameEnd = nameStart;
	}

	//System.err.println("Return arguments of:\n" + initArgs);

	return initArgs;
    }
}
