/*  ----------------------------------------------------------------<Prolog>-
    Name:       com.imatix.ServletRunner.WebServerConnection
    Title:      Interface: Connection to web server
    Package:    iMatix Servlet Runner

    Written:    99/01/11  Ewen McNeill <ewen@imatix.com>
    Revised:    99/01/19

    Copyright:  Copyright (c) 1991-99 iMatix Corporation
    License:    This is free software; you can redistribute it and/or modify
                it under the terms of the SMT License Agreement as provided
                in the file LICENSE.TXT.  This software is distributed in
                the hope that it will be useful, but without any warranty.
 ------------------------------------------------------------------</Prolog>-*/

package com.imatix.ServletRunner;

import java.util.Hashtable;

/**
    This interface encapsulates the required functionality for a
    persistent connection to a web server to retrieve and respond to
    requests for servlets to run.  Classes conforming to this interface
    can be used by the iMatix servlet runner as a means of communicating
    with the web server.
 */
interface WebServerConnection
{
    /**
        Establish connection to the web server to indicate willingness
	to handle connections for the given applicationName on a given
	virtualHost.  The virtualHost may be unspecified (null), which
	means that the applicationName will apply on all hosts.  The
	accessKey may or may not be required depending on the type of
	connection being made.

	If this function retruns false, an indicative error message can
	be retrieved with getErrorMessage().
     */
    public boolean connectToWebServer (String applicationName,
                                       String virtualHost,
				       String accessKey);

    /**
        Wait for new requests, and accept the details of the request.
	The actual details can be retrieved with other functions, such
	as getRequestContext(), and getRequestDate(), to retrieve the
	equivilent of the CGI environment, and the posted data (if any).

	The exception NotConnectedToWebServer will be thrown if the
	connection to the web server has not been established yet, or
	the connection has been lost.
     */
    public boolean waitForRequest () throws NotConnectedToWebServer;

    /**
        Retrieve the context of the request.  This returns a hash table
	consisting of (at least) the elements that are available in the
	environment to a CGI program when run from a CGI 1.1 compliant
	server.

	The exception NotConnectedToWebServer will be thrown if the
	connection to the web server has not been established yet, or
	the connection has been lost.
	The exception NoActiveRequest will be thrown if there is not
	an open request at the time this function is called.
      */
    public Hashtable getRequestContext () throws NoActiveRequest;

    /**
        Retrieve the input data from the request, if any.  Returns 
	null if there is no input data to be processed.

	The exception NotConnectedToWebServer will be thrown if the
	connection to the web server has not been established yet, or
	the connection has been lost.
	The exception NoActiveRequest will be thrown if there is not
	an open request at the time this function is called.
     */
    public byte[] getRequestInput () throws NoActiveRequest;

    /** 
        Send response to the web server.  The response is provided as
	an array of bytes to send to the webserver.

	The exception NotConnectedToWebServer will be thrown if the
	connection to the web server has not been established yet, or
	the connection has been lost.
	The exception NoActiveRequest will be thrown if there is not
	an open request at the time this function is called.
     */
    public boolean sendResponse (byte[] response);
    public boolean sendResponse (byte[] response, int startAt);
    public boolean sendResponse (byte[] response, int startAt, int endAt);

    /**
        Retrieve error message (if any).  The error message is useful 
	only if an error has occured, otherwise it contains an
	innocent string.
     */
    public String getErrorMessage();

    /** 
        Disconnect from the web server.  This can be done while a
	request is in progress, but will result in that request being
	abandoned.  If there is no active connection to the web server,
	this method will be a noop.
     */
    public void disconnectFromWebServer();
}
